package inf1010.pusle.adventure;

import java.util.EnumMap;

/**
 * A character in the game (NPC or the player).
 */
class Character extends Entity implements Comparable<Character> {

    /** The character's name. */
    final String name;
    /** Base stats. */
    final int maxHp, baseAttack, baseDefense;
    /** Modified stats. */
    int hp, attack, defense;
    /** The room the character is in. */
    Room location;
    /** Items carried by the character. */
    SinglyLinkedList<Item> items = new SinglyLinkedList<Item>();
    /** Equipment worn/wielded. */
    EnumMap<BodyPart, Equipment> equipment = new EnumMap<BodyPart,
	Equipment>(BodyPart.class);
	
    /**
     * Create a new character with the given attributes.
     *
     * @param name The character's name.
     * @param desc A description of the character.
     * @param maxHp The character's maximum HP.
     * @param baseAttack This character's base attack.
     * @param baseDefense This character's base defense.
     * @param loc The spawn room for this character.
     */
    public Character(String name, String desc,
		     int maxHp, int baseAttack, int baseDefense, Room loc) {
	super(desc);
	this.name = name;
	this.maxHp = maxHp;
	this.baseAttack = baseAttack;
	this.baseDefense = baseDefense;
	this.location = loc;
	this.hp = maxHp;
	this.attack = baseAttack;
	this.defense = baseDefense;
    }
	
    /**
     * Try to kill the given opponent.
     *
     * @param opponent The character the player wants to kill.
     * @return true if the opponent was killed,
     *         false if this character was killed.
     */
    public boolean kill(Character opponent) {
	while (this.hp > 0 && opponent.hp > 0) {

	    int dmg = (int)(Math.random() * attack
			    - Math.random() * opponent.defense);
	    if (dmg > 0) {
		System.out.printf("You hit %s.%n", opponent.name);
		opponent.hp -= dmg;
	    } else {
		System.out.printf("%s dodges your swing.%n", opponent.name);
	    }
			
	    if (opponent.hp <= 0)
		break;

	    dmg = (int)(Math.random() * opponent.attack
			    - Math.random() * defense);

	    if (dmg > 0) {
		System.out.printf("%s hits you.%n", opponent.name);
		hp -= dmg;
	    } else {
		System.out.printf("You dodge %s's swing.%n", opponent.name);
	    }

	    try {
		Thread.sleep(1000);
	    } catch (InterruptedException e) { }
	}
		
	if (hp <= 0) {
	    System.out.printf("You die a horrible death at the hands of %s.%n", opponent.name);
	    return false;
	}

	System.out.printf("You killed %s.%n", opponent.name);

	if (opponent.items.isEmpty())
	    System.out.printf("The bastard wasn't carrying anything.%n");

	for (Item it : opponent.items) {
	    location.items.add(it);
	    System.out.printf("%s drops a%s %s.%n",
			      opponent.name,
			      (it.startsWithVowel() ? "n" : ""),
			      it.name);
	}
	location.npcs.remove(opponent);
	return true;
    }
	
    /**
     * Compares two characters' names.
     *
     * @param c The character to compare names with.
     * @return 0 if the names are the same,
     *		   a negative number if it is lexically lower
     * 		   and a positive number if it is lexically higher.
     */
    public int compareTo(Character c) {
	if (c == null)
	    return -1;

	return name.toLowerCase().compareTo(c.name.toLowerCase());
    }

    /**
     * Prints a message evaluating how tough a character would be to beat.
     * This algorithm could be completely wrong.
     *
     * @param c The character to consider.
     */
    public void consider(Character c) {
	int hpCmp = hp - c.hp;
	int atkCmp = attack - c.attack;
	int defCmp = defense - c.defense;
	int cmp = atkCmp + defCmp + hpCmp;

	if (cmp > 100) {
	    System.out.printf("%s looks like a wuss!%n", c.name);
	} else if (cmp > 50) {
	    System.out.printf("%s doesn't look like much of a fight.%n", c.name);
	} else if (cmp > 10) {
	    System.out.printf("You should be able to beat %s.%n", c.name);
	} else if (cmp > -10) {
	    System.out.printf("%s looks like an even fight.%n", c.name);
	} else if (cmp > -50) {
	    System.out.printf("You probably can't beat %s.%n", c.name);
	} else if (cmp > -100) {
	    System.out.printf("%s looks like a hard fight.%n", c.name);
	} else {
	    System.out.printf("%s looks like a tough guy!%n", c.name);
	}
    }

    /**
     * Drops an item in the inventory on the ground.
     *
     * @param it The item to drop.
     */
    public void drop(Item it) {
	it = items.get(it);
	
	if (it == null) {
	    return;
	}

	// If the player wants to drop equipment being used,
	// it has to be removed first.
	if (it instanceof Equipment) {
	    Equipment e = (Equipment)it;
	    Equipment e2 = equipment.get(e.bp);
	    if (e2 != null && e.compareTo(e2) == 0)
		unequip(e, false);
	}

	System.out.printf("You drop the %s.%n", it.name);
	location.items.add(it);
	items.remove(it);
    }

    /**
     * Equips a piece of equipment.
     *
     * @param e The piece of equipment to equip.
     * @param silent Whether to print equipping info.
     */
    public void equip(Equipment e, boolean silent) {
	Equipment toEquip = (Equipment)items.get(e);
	if (toEquip == null) {
	    if (!silent)
		System.out.println("You cannot equip something you do not have!");
	    return;
	}

	Equipment oldEquipment = equipment.put(toEquip.bp, toEquip);
	if (oldEquipment != null) {
	    System.out.println("Unequipping crap!");
	    unequip(oldEquipment, silent);
	}

	if (!silent)
	    System.out.printf("You equip the %s.%n", toEquip.name);
	attack += toEquip.attackBonus;
	defense += toEquip.defenseBonus;
	equipment.put(toEquip.bp, toEquip);
    }

    /**
     * Unequips a piece of equipment.
     *
     * @param e The equipment to unequip.
     * @param silent Whether to print unequip messages.
     */
    public void unequip(Equipment e, boolean silent) {
	Equipment toUnequip = equipment.get(e.bp);

	if (toUnequip == null || e.compareTo(toUnequip) != 0) {
	    if (!silent)
		System.out.printf("You're not wearing that.%n");
	    return;
	}

	if (!silent)
	    System.out.printf("You unequip the %s.%n", toUnequip.name);
	attack -= toUnequip.attackBonus;
	defense -= toUnequip.defenseBonus;
	equipment.remove(toUnequip.bp);
    }
}

