package inf1010.pusle.adventure;


import java.util.Map;
import java.util.HashMap;

import java.io.InputStream;
import java.io.IOException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * A simple map loader utility class.
 */
class MapLoader {

    /**
     * Disable instancing.
     */
    private MapLoader() { }

    /**
     * Loads a map and returns the spawn room.
     */
    public static Room loadMap() throws IOException,
					ParserConfigurationException,
					SAXException {
	InputStream is = MapLoader.class.getResourceAsStream("/res/map.xml");
	if (is == null) {
	    throw new IOException("Could not load res/map.xml.");
	}

	Document mapDoc = DocumentBuilderFactory.newInstance().
	    newDocumentBuilder().parse(is);
	is.close();

	NodeList roomNodes = mapDoc.getElementsByTagName("room");
	Map<Integer, Room> rooms = new HashMap<Integer, Room>();
	int id;
	String desc = null;
	Node node;
	NodeList roomChildren;
	for (int i = 0; i < roomNodes.getLength(); i++) {
	    node = roomNodes.item(i);
	    id = Integer.parseInt(node.getAttributes().
				  getNamedItem("id").getNodeValue());

	    roomChildren = node.getChildNodes();
	    for (int j = 0; j < roomChildren.getLength(); j++) {
		if (roomChildren.item(j).getNodeName().equals("desc"))
		    desc = roomChildren.item(j).getFirstChild().getNodeValue();
	    }

	    rooms.put(id, new Room(desc, id));
	}

	for (int i = 0; i < roomNodes.getLength(); i++) {
	    node = roomNodes.item(i);
	    id = Integer.parseInt(node.getAttributes().
				  getNamedItem("id").getNodeValue());
	    Room room = rooms.get(id);

	    roomChildren = roomNodes.item(i).getChildNodes();
	    for (int j = 0; j < roomChildren.getLength(); j++) {
		Node n = roomChildren.item(j);
		Direction d = Direction.getDirection(n.getNodeName());
		if (d != null) {
		    int linkId = Integer.parseInt(n.getAttributes().
						  getNamedItem("id").
						  getNodeValue());
		    room.addExit(d, rooms.get(linkId));
		} else if (n.getNodeName().equals("npcs")) {
		    loadNPCs(room, n);
		} else if (n.getNodeName().equals("items")) {
		    loadItems(room, n);
		}
	    }
	}

	node = mapDoc.getElementsByTagName("spawn").item(0);
	id = Integer.parseInt(node.getAttributes().
			      getNamedItem("id").getNodeValue());

	return rooms.get(id);
    }

    /**
     * Loads NPCs for a room from the npcs node.
     *
     * @param r The room we're in.
     * @param npcsNode The npcs node containing a list of npcs.
     */
    private static void loadNPCs(Room r, Node npcsNode) {
	NodeList npcs = npcsNode.getChildNodes();
	NodeList npcText;
	NamedNodeMap attrs;
	int hp, atk, def;
	String desc = null;
	String name = null;

	for (int i = 0; i < npcs.getLength(); i++) {
	    if (!npcs.item(i).getNodeName().equals("npc")) {
		continue;
	    }

	    attrs = npcs.item(i).getAttributes();
	    hp = Integer.parseInt(attrs.getNamedItem("hp").getNodeValue());
	    atk = Integer.parseInt(attrs.getNamedItem("atk").getNodeValue());
	    def = Integer.parseInt(attrs.getNamedItem("def").getNodeValue());

	    npcText = npcs.item(i).getChildNodes();
	    for (int j = 0; j < npcText.getLength(); j++) {
		if (npcText.item(j).getNodeName().equals("name")) {
		    name = npcText.item(j).getFirstChild().getNodeValue();
		} else if (npcText.item(j).getNodeName().equals("desc")) {
		    desc = npcText.item(j).getFirstChild().getNodeValue();
		}
	    }

	    Character c = new Character(name, desc, hp, atk, def, r);
	    for (int j = 0; j < npcText.getLength(); j++) {
		if (npcText.item(j).getNodeName().equals("items")) {
		    loadItems(c, npcText.item(j));
		}
	    }
	    r.npcs.add(c);
	}
    }

    /**
     * Handles loading of items.
     *
     * Both equipment and useless junk items.
     */
    private static void loadItems(Entity e,
				  Node itemsNode) {
	NodeList items = itemsNode.getChildNodes();
	SinglyLinkedList<Item> itemList = null;
	NodeList itemText;
	NamedNodeMap eqAttrs;
	int atkBoost, defBoost;
	BodyPart bp = null;
	boolean isEquipped = false;
	String desc = null;
	String name = null;
	
	if (e instanceof Room) {
	    itemList = ((Room) e).items;
	} else if (e instanceof Character) {
	    itemList = ((Character) e).items;
	} else {
	    // Here we could print an error message or throw an Exception.
	    // But we're being a bit lazy for now.
	    return;
	}

	for (int i = 0; i < items.getLength(); i++) {
	    if (!items.item(i).getNodeName().equals("item")
		&& !items.item(i).getNodeName().equals("equip")) {
		continue;
	    }

	    itemText = items.item(i).getChildNodes();
	    for (int j = 0; j < itemText.getLength(); j++) {
		if (itemText.item(j).getNodeName().equals("name")) {
		    name = itemText.item(j).getFirstChild().getNodeValue();
		} else if (itemText.item(j).getNodeName().equals("desc")) {
		    desc = itemText.item(j).getFirstChild().getNodeValue();
		}
	    }

	    if (items.item(i).getNodeName().equals("item")) {
		itemList.add(new Item(name, desc));
	    } else {
		eqAttrs = items.item(i).getAttributes();
		bp = BodyPart.getBodyPart(eqAttrs.
					  getNamedItem("bp").
					  getNodeValue());
		atkBoost = Integer.parseInt(eqAttrs.
					    getNamedItem("atk").
					    getNodeValue());
		defBoost = Integer.parseInt(eqAttrs.
					    getNamedItem("def").
					    getNodeValue());
		isEquipped = Boolean.parseBoolean(eqAttrs.
						  getNamedItem("equipped").
						  getNodeValue());
		Equipment eq = new Equipment(name, desc,
					     atkBoost, defBoost, bp);
		itemList.add(eq);
		if (isEquipped && e instanceof Character) {
		    ((Character) e).equip(eq, true);
		}
	    }
	}
    }
}

